from typing import List, Union

from django.contrib import admin

from edc_visit_tracking.stubs import TSubjectVisitModelStub


class CrfModelAdminMixin:

    """ModelAdmin subclass for models with a ForeignKey to your
    visit model(s).
    """

    date_hierarchy = "report_datetime"

    def visit_reason(self, obj=None):
        return getattr(obj, self.visit_model_attr).reason

    def visit_code(self, obj=None):
        return getattr(obj, self.visit_model_attr).appointment.visit_code

    def subject_identifier(self, obj=None):
        return getattr(obj, self.visit_model_attr).subject_identifier

    def get_list_display(self: Union["CrfModelAdminMixin", admin.ModelAdmin], request) -> List:
        list_display = super().get_list_display(request)  # type: ignore
        fields = [self.visit_code, self.visit_reason]
        fields_first = [self.subject_identifier, "report_datetime"]
        list_display = list(list_display)
        try:
            list_display.remove("__str__")
        except ValueError:
            pass
        list_display = (
            [f for f in fields_first if f not in list_display]
            + list_display
            + [f for f in fields if f not in list_display]
        )
        return list_display

    def get_search_fields(self, request) -> List:
        search_fields = super().get_search_fields(request)  # type: ignore
        fields = [f"{self.visit_model_attr}__appointment__subject_identifier"]
        search_fields = [f for f in fields if f not in search_fields] + list(search_fields)
        return search_fields

    def get_list_filter(self: Union["CrfModelAdminMixin", admin.ModelAdmin], request) -> List:
        list_filter = super().get_list_filter(request)  # type: ignore
        fields = [
            f"{self.visit_model_attr}__report_datetime",
            f"{self.visit_model_attr}__appointment__visit_code",
            f"{self.visit_model_attr}__reason",
            f"{self.visit_model_attr}__appointment__appt_status",
        ]
        list_filter = [f for f in list_filter if f not in fields] + fields
        return list_filter

    @property
    def visit_model(self: admin.ModelAdmin) -> TSubjectVisitModelStub:
        return self.model.visit_model_cls()

    @property
    def visit_model_attr(self: admin.ModelAdmin) -> str:
        return self.model.visit_model_attr()

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        db = kwargs.get("using")
        if db_field.name == self.visit_model_attr and request.GET.get(self.visit_model_attr):
            if request.GET.get(self.visit_model_attr):
                kwargs["queryset"] = self.visit_model._default_manager.using(db).filter(
                    id__exact=request.GET.get(self.visit_model_attr)
                )
            else:
                kwargs["queryset"] = self.visit_model._default_manager.none()
        return super().formfield_for_foreignkey(db_field, request, **kwargs)  # type: ignore

    def get_readonly_fields(self, request, obj=None) -> List:
        readonly_fields = super().get_readonly_fields(request, obj=obj)  # type: ignore
        if (
            not request.GET.get(self.visit_model_attr)
            and self.visit_model_attr not in readonly_fields
        ):
            readonly_fields = list(readonly_fields)
            readonly_fields.append(self.visit_model_attr)
        return readonly_fields



from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_appointment.managers import AppointmentDeleteError
from edc_appointment.models import Appointment
from edc_facility.import_holidays import import_holidays
from edc_reference import site_reference_configs
from edc_utils import get_utcnow
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from edc_visit_tracking.constants import SCHEDULED, UNSCHEDULED
from edc_visit_tracking.model_mixins import PreviousVisitError
from edc_visit_tracking.visit_sequence import VisitSequence, VisitSequenceError

from ..helper import Helper
from ..models import SubjectVisit
from ..visit_schedule import visit_schedule1, visit_schedule2


class DisabledVisitSequence(VisitSequence):
    def enforce_sequence(self):
        return None


class TestPreviousVisit(TestCase):
    helper_cls = Helper

    @classmethod
    def setUpClass(cls):
        import_holidays()
        return super().setUpClass()

    def setUp(self):
        SubjectVisit.visit_sequence_cls = VisitSequence
        self.subject_identifier = "12345"
        self.helper = self.helper_cls(subject_identifier=self.subject_identifier)
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule=visit_schedule1)
        site_visit_schedules.register(visit_schedule=visit_schedule2)
        site_reference_configs.register_from_visit_schedule(
            visit_models={"edc_appointment.appointment": "edc_visit_tracking.subjectvisit"}
        )
        self.helper.consent_and_put_on_schedule()

    def tearDown(self):
        SubjectVisit.visit_sequence_cls = VisitSequence

    def test_visit_sequence_enforcer_on_first_visit_in_sequence(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        SubjectVisit.visit_sequence_cls = DisabledVisitSequence
        visit = SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )
        visit_sequence = VisitSequence(appointment=visit.appointment)
        try:
            visit_sequence.enforce_sequence()
        except VisitSequenceError as e:
            self.fail(f"VisitSequenceError unexpectedly raised. Got '{e}'")

    def test_visit_sequence_enforcer_without_first_visit_in_sequence(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        SubjectVisit.visit_sequence_cls = DisabledVisitSequence
        visit = SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )
        visit_sequence = VisitSequence(appointment=visit.appointment)
        self.assertRaises(VisitSequenceError, visit_sequence.enforce_sequence)

    def test_requires_previous_visit_thru_model(self):
        """Asserts requires previous visit to exist on create."""
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )
        self.assertRaises(
            PreviousVisitError,
            SubjectVisit.objects.create,
            appointment=appointments[2],
            report_datetime=get_utcnow() - relativedelta(months=8),
            reason=SCHEDULED,
        )
        SubjectVisit.objects.create(
            appointment=appointments[1],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )
        self.assertRaises(
            PreviousVisitError,
            SubjectVisit.objects.create,
            appointment=appointments[3],
            report_datetime=get_utcnow() - relativedelta(months=8),
            reason=SCHEDULED,
        )

    def test_requires_previous_visit_thru_model2(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        opts = appointments[1].__dict__
        opts.pop("_state")
        opts.pop("id")
        opts.pop("created")
        opts.pop("modified")

        opts["visit_code_sequence"] = 1
        Appointment.objects.create(**opts)
        opts["visit_code_sequence"] = 2

        SubjectVisit.objects.create(
            appointment=appointments[0],
            report_datetime=get_utcnow() - relativedelta(months=10),
            reason=SCHEDULED,
        )

        self.assertRaises(
            PreviousVisitError,
            SubjectVisit.objects.create,
            appointment=appointments[2],
            report_datetime=get_utcnow() - relativedelta(months=8),
        )

    def test_previous_appointment(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        visit_sequence = VisitSequence(appointment=appointments[0])
        self.assertIsNone(visit_sequence.previous_appointment)
        visit_sequence = VisitSequence(appointment=appointments[1])
        self.assertEqual(visit_sequence.previous_appointment, appointments[0])
        visit_sequence = VisitSequence(appointment=appointments[2])
        self.assertEqual(visit_sequence.previous_appointment, appointments[1])

    def test_previous_appointment_with_unscheduled(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        # insert some unscheduled appointments
        opts = appointments[1].__dict__
        opts.pop("_state")
        opts.pop("id")
        opts.pop("created")
        opts.pop("modified")
        opts["visit_code_sequence"] = 1
        Appointment.objects.create(**opts)
        opts["visit_code_sequence"] = 2
        Appointment.objects.create(**opts)

        visit_sequence = VisitSequence(appointment=appointments[0])
        self.assertIsNone(visit_sequence.previous_appointment)
        for i in range(0, Appointment.objects.all().count() - 1):
            visit_sequence = VisitSequence(appointment=appointments[i + 1])
            self.assertEqual(visit_sequence.previous_appointment, appointments[i])

    def test_previous_appointment_broken_sequence1(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        self.assertRaises(AppointmentDeleteError, appointments[1].delete)

    def test_previous_appointment_broken_sequence2(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")

        opts = appointments[1].__dict__
        opts.pop("_state")
        opts.pop("id")
        opts.pop("created")
        opts.pop("modified")
        opts["visit_code_sequence"] = 2
        Appointment.objects.create(**opts)

        visit_sequence = VisitSequence(appointment=appointments[0])
        self.assertIsNone(visit_sequence.previous_appointment)
        visit_sequence = VisitSequence(appointment=appointments[1])
        self.assertEqual(visit_sequence.previous_appointment, appointments[0])

        visit_sequence = VisitSequence(appointment=appointments[2])
        self.assertRaises(VisitSequenceError, getattr, visit_sequence, "previous_appointment")

        visit_sequence = VisitSequence(appointment=appointments[3])
        self.assertEqual(visit_sequence.previous_appointment, appointments[2])

    def test_previous_visit(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        for index, appointment in enumerate(appointments):
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED,
            )

    def test_previous_visit2(self):
        appointments = Appointment.objects.all().order_by("timepoint", "visit_code_sequence")
        opts = appointments[1].__dict__
        opts.pop("_state")
        opts.pop("id")
        opts.pop("created")
        opts.pop("modified")
        opts["visit_code_sequence"] = 1
        Appointment.objects.create(**opts)
        opts["visit_code_sequence"] = 2
        Appointment.objects.create(**opts)

        for index, appointment in enumerate(appointments):
            SubjectVisit.objects.create(
                appointment=appointment,
                report_datetime=get_utcnow() - relativedelta(months=10 - index),
                reason=SCHEDULED if appointment.visit_code_sequence == 0 else UNSCHEDULED,
            )



from .caretaker_fields_mixin import CaretakerFieldsMixin
from .crfs import VisitTrackingCrfModelMixin
from .previous_visit_model_mixin import PreviousVisitError, PreviousVisitModelMixin
from .subject_visit_missed_model_mixin import SubjectVisitMissedModelMixin
from .visit_model_mixin import VisitModelFieldsMixin, VisitModelMixin



# Generated by Django 3.1.8 on 2021-04-23 11:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edc_visit_tracking', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='subjectvisitmissedreasons',
            name='id',
            field=models.AutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='visitreasons',
            name='id',
            field=models.AutoField(primary_key=True, serialize=False),
        ),
    ]



from typing import Any, Optional

from django import forms


def get_subject_visit(modelform: Any, visit_model_attr: Optional[str]):
    if visit_model_attr not in modelform.cleaned_data:
        subject_visit = getattr(modelform.instance, visit_model_attr, None)
        if not subject_visit:
            raise forms.ValidationError(f"Field `{visit_model_attr}` is required (2).")
    else:
        subject_visit = modelform.cleaned_data.get(visit_model_attr)
    return subject_visit



from copy import deepcopy
from typing import List, Optional
from zoneinfo import ZoneInfo

from arrow import Arrow
from django import forms
from django.conf import settings
from edc_appointment.constants import MISSED_APPT
from edc_appointment.form_validators import WindowPeriodFormValidatorMixin
from edc_constants.constants import OTHER
from edc_form_validators import INVALID_ERROR, REQUIRED_ERROR, FormValidator
from edc_metadata.constants import KEYED
from edc_metadata.models import CrfMetadata, RequisitionMetadata
from edc_utils import formatted_datetime

from edc_visit_tracking.utils import get_subject_visit_missed_model_cls

from ..constants import MISSED_VISIT, SCHEDULED, UNSCHEDULED
from ..visit_sequence import VisitSequence, VisitSequenceError

EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED = getattr(
    settings, "EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED", False
)


class VisitFormValidator(WindowPeriodFormValidatorMixin, FormValidator):

    """Form validator for visit all models (e.g. subject_visit).

    See also `report_datetime` checks in the
    `VisitTrackingModelFormMixin`.
    """

    visit_sequence_cls = VisitSequence
    validate_missed_visit_reason = True
    validate_unscheduled_visit_reason = True

    def _clean(self) -> None:
        self.clean_defaults()
        super()._clean()

    def clean_defaults(self) -> None:
        if not self.cleaned_data.get("appointment"):
            raise forms.ValidationError(
                {"appointment": "This field is required"}, code=REQUIRED_ERROR
            )

        self.validate_visit_datetime_not_before_appointment()

        self.validate_visit_datetime_in_window_period()

        self.validate_visits_completed_in_order()

        self.validate_visit_code_sequence_and_reason()

        self.validate_visit_reason()

        self.required_if(OTHER, field="info_source", field_required="info_source_other")

    def validate_visit_datetime_not_before_appointment(
        self,
    ) -> None:
        """Asserts the report_datetime is not before the
        appt_datetime.
        """
        if report_datetime := self.cleaned_data.get("report_datetime"):
            tz = ZoneInfo(settings.TIME_ZONE)
            appt_datetime_local = Arrow.fromdatetime(
                self.cleaned_data.get("appointment").appt_datetime
            ).to(tz)
            if report_datetime.date() < appt_datetime_local.date():
                appt_datetime_str = formatted_datetime(
                    appt_datetime_local, format_as_date=True
                )
                self.raise_validation_error(
                    {
                        "report_datetime": (
                            "Invalid. Cannot be before appointment date. "
                            f"Got appointment date {appt_datetime_str}"
                        )
                    },
                    INVALID_ERROR,
                )

    def validate_visit_datetime_in_window_period(self):
        """Asserts the report_datetime is within the visits lower and
        upper boundaries of the visit_schedule.schdule.visit.

        See also `edc_visit_schedule`.
        """
        if self.cleaned_data.get("report_datetime"):
            super().validate_visit_datetime_in_window_period(
                self.cleaned_data.get("appointment"),
                self.cleaned_data.get("report_datetime"),
                "report_datetime",
            )

    def validate_visits_completed_in_order(self):
        """Asserts visits are completed in order."""
        visit_sequence = self.visit_sequence_cls(
            appointment=self.cleaned_data.get("appointment")
        )
        try:
            visit_sequence.enforce_sequence()
        except VisitSequenceError as e:
            raise forms.ValidationError(e, code=INVALID_ERROR)

    def validate_visit_code_sequence_and_reason(self) -> None:
        """Asserts the `reason` makes sense relative to the
        visit_code_sequence coming from the appointment.
        """
        appointment = self.cleaned_data.get("appointment")
        reason = self.cleaned_data.get("reason")
        if appointment:
            if not appointment.visit_code_sequence and reason == UNSCHEDULED:
                raise forms.ValidationError(
                    {"reason": "Invalid. This is not an unscheduled visit. See appointment."},
                    code=INVALID_ERROR,
                )
            if (
                appointment.visit_code_sequence
                and reason != UNSCHEDULED
                and EDC_VISIT_TRACKING_ALLOW_MISSED_UNSCHEDULED is False
            ):
                raise forms.ValidationError(
                    {"reason": "Invalid. This is an unscheduled visit. See appointment."},
                    code=INVALID_ERROR,
                )
            # raise if CRF metadata exist
            if reason == MISSED_VISIT and self.metadata_exists_for(
                entry_status=KEYED,
                exclude_models=[get_subject_visit_missed_model_cls()._meta.label_lower],
            ):
                raise forms.ValidationError(
                    {"reason": "Invalid. Some CRF data has already been submitted."},
                    code=INVALID_ERROR,
                )
            # raise if SubjectVisitMissed CRF metadata exist
            if reason in [UNSCHEDULED, SCHEDULED] and self.metadata_exists_for(
                entry_status=KEYED,
                filter_models=[get_subject_visit_missed_model_cls()._meta.label_lower],
            ):
                raise forms.ValidationError(
                    {"reason": "Invalid. A missed visit report has already been submitted."},
                    code=INVALID_ERROR,
                )

    def validate_visit_reason(self):
        """Asserts that reason=missed if appointment is missed"""
        if (
            self.cleaned_data.get("appointment").appt_timing == MISSED_APPT
            and self.cleaned_data.get("reason") != MISSED_VISIT
        ):
            self.raise_validation_error(
                {"reason": "Invalid. This is a missed visit. See appointment"}, INVALID_ERROR
            )

        if self.validate_missed_visit_reason:
            self.required_if(MISSED_VISIT, field="reason", field_required="reason_missed")

            self.required_if(
                OTHER, field="reason_missed", field_required="reason_missed_other"
            )

        if self.validate_unscheduled_visit_reason:
            if "reason_unscheduled" in self.cleaned_data:
                self.applicable_if(
                    UNSCHEDULED, field="reason", field_applicable="reason_unscheduled"
                )

                self.required_if(
                    OTHER,
                    field="reason_unscheduled",
                    field_required="reason_unscheduled_other",
                )

    def metadata_exists_for(
        self,
        entry_status: str = None,
        filter_models: Optional[List[str]] = None,
        exclude_models: Optional[List[str]] = None,
    ) -> bool:
        """Returns True if metadata exists for this visit for
        the given entry_status.
        """
        exclude_opts: dict = {}
        filter_opts = deepcopy(self.crf_filter_options)
        filter_opts.update(entry_status=entry_status or KEYED)
        if filter_models:
            filter_opts.update(model__in=filter_models)
        if exclude_models:
            exclude_opts.update(model__in=exclude_models)
        return (
            CrfMetadata.objects.filter(**filter_opts).exclude(**exclude_opts).count()
            + RequisitionMetadata.objects.filter(**filter_opts).exclude(**exclude_opts).count()
        )

    @property
    def crf_filter_options(self) -> dict:
        """Returns a dictionary of `filter` options when querying
        models CrfMetadata / RequisitionMetadata.
        """
        appointment = self.cleaned_data.get("appointment")
        return dict(
            subject_identifier=appointment.subject_identifier,
            visit_code=appointment.visit_code,
            visit_code_sequence=appointment.visit_code_sequence,
            visit_schedule_name=appointment.visit_schedule_name,
            schedule_name=appointment.schedule_name,
            entry_status=KEYED,
        )
